﻿using System;
using System.Collections.Generic;
using System.Windows.Media.Media3D;

namespace TestProject
{
    /// <summary>
    /// 삼각형
    /// </summary>
    public class Triangle
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// 포인트 배열
        /// </summary>
        public Point3D[] PointArray;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Triangle(pointArray)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="pointArray">포인트 배열</param>
        public Triangle(params Point3D[] pointArray)
        {
            PointArray = pointArray;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 세분화 하기 - Subdivide(triangleList, centerPoint, radius)

        /// <summary>
        /// 세분화 하기
        /// </summary>
        /// <param name="triangleList">삼각형 리스트</param>
        /// <param name="centerPoint">중심 포인트</param>
        /// <param name="radius">반지름</param>
        public void Subdivide(List<Triangle> triangleList, Point3D centerPoint, double radius)
        {
            Vector3D v10 = PointArray[1] - PointArray[0];
            Vector3D v20 = PointArray[2] - PointArray[0];
            Vector3D v21 = PointArray[2] - PointArray[1];

            Point3D point1 = PointArray[0] + v10 * 1.0 / 3.0;
            Point3D point2 = PointArray[0] + v20 * 1.0 / 3.0;
            Point3D point3 = PointArray[0] + v10 * 2.0 / 3.0;
            Point3D point4 = PointArray[0] + v10 * 2.0 / 3.0 + v21 * 1.0 / 3.0;
            Point3D point5 = PointArray[0] + v20 * 2.0 / 3.0;
            Point3D point6 = PointArray[1] + v21 * 1.0 / 3.0;
            Point3D point7 = PointArray[1] + v21 * 2.0 / 3.0;

            NormalizePoint(ref point1, centerPoint, radius);
            NormalizePoint(ref point2, centerPoint, radius);
            NormalizePoint(ref point3, centerPoint, radius);
            NormalizePoint(ref point4, centerPoint, radius);
            NormalizePoint(ref point5, centerPoint, radius);
            NormalizePoint(ref point6, centerPoint, radius);
            NormalizePoint(ref point7, centerPoint, radius);

            triangleList.Add(new Triangle(PointArray[0], point1       , point2       ));
            triangleList.Add(new Triangle(point1       , point3       , point4       ));
            triangleList.Add(new Triangle(point1       , point4       , point2       ));
            triangleList.Add(new Triangle(point2       , point4       , point5       ));
            triangleList.Add(new Triangle(point3       , PointArray[1], point6       ));
            triangleList.Add(new Triangle(point3       , point6       , point4       ));
            triangleList.Add(new Triangle(point4       , point6       , point7       ));
            triangleList.Add(new Triangle(point4       , point7       , point5       ));
            triangleList.Add(new Triangle(point5       , point7       , PointArray[2]));
        }

        #endregion
        #region 각도 배열 구하기 - GetAngleArray()

        /// <summary>
        /// 각도 배열 구하기
        /// </summary>
        /// <returns></returns>
        public double[] GetAngleArray()
        {
            Vector3D v10 = PointArray[1] - PointArray[0];
            Vector3D v21 = PointArray[2] - PointArray[1];
            Vector3D v02 = PointArray[0] - PointArray[2];

            v10.Normalize();
            v21.Normalize();
            v02.Normalize();

            double angle0 = Math.Acos(Vector3D.DotProduct(v10, -v02)) * 180.0 / Math.PI;
            double angle1 = Math.Acos(Vector3D.DotProduct(v21, -v10)) * 180.0 / Math.PI;
            double angle2 = Math.Acos(Vector3D.DotProduct(v02, -v21)) * 180.0 / Math.PI;

            double[] angleArray = new double[] { angle0, angle1, angle2 };

            Array.Sort(angleArray);

            return angleArray;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 포인트 정규화 하기 - NormalizePoint(point, centerPoint, distance)

        /// <summary>
        /// 포인트 정규화 하기
        /// </summary>
        /// <param name="point">포인트</param>
        /// <param name="centerPoint">중심 포인트</param>
        /// <param name="distance">거리</param>
        private void NormalizePoint(ref Point3D point, Point3D centerPoint, double distance)
        {
            Vector3D vector = point - centerPoint;

            point = centerPoint + vector / vector.Length * distance;
        }

        #endregion
    }
}